import contextlib
import logging
import sys

import celery
from celery.utils.log import get_task_logger
from celery import signals

PY2 = sys.version_info[0] == 2


@signals.celeryd_init.connect
def configure_worker(conf=None, **kwargs):
    conf.setdefault('CUBICWEB_CELERYTASK_REDIS_URL',
                    'redis://localhost:6379/0')


@signals.task_failure.connect
def log_exception(**kwargs):
    logger = logging.getLogger("celery.task")
    if PY2:
        einfo = str(kwargs['einfo']).decode('utf8')
    else:
        einfo = str(kwargs['einfo'])
    logger.critical(u"unhandled exception:\n%s", einfo)


@signals.celeryd_after_setup.connect
def setup_cubicweb_logging(conf=None, **kwargs):
    """
    Set parent to "celery.task" for all instantiated logger names starting with
    "cube" or "cubicweb"
    """
    logall = conf.get('CUBICWEB_CELERYTASK_LOG_ALL', False)
    for logname in logging.root.manager.loggerDict.keys():
        if logname == 'celery' or logname.startswith('celery.'):
            # Logger name 'celery' is reserved
            continue
        if(logall
           or logname.startswith('cubes')
           or logname.startswith('cubicweb')):
            get_task_logger(logname)


@contextlib.contextmanager
def redirect_stdouts(logger):
    old_outs = sys.stdout, sys.stderr
    try:
        app = celery.current_app
        rlevel = app.conf.CELERY_REDIRECT_STDOUTS_LEVEL
        app.log.redirect_stdouts_to_logger(logger, rlevel)
        yield
    finally:
        sys.stdout, sys.stderr = old_outs



"""Helpers for managing logging in redis in cubicweb-celerytask workers

Add this module 'cw_celerytask_helpers.redislogger' to CELERY_IMPORTS
"""
from __future__ import absolute_import

import json
import datetime
import logging

from celery._state import get_current_task
from celery import signals

from .redisutils import get_log_key, get_redis_client, LOG_KEY_PREFIX
from .redisutils import get_task_logs, flush_task_logs  # noqa


class UnknownTaskId(Exception):
    pass


@signals.celeryd_after_setup.connect
def setup_redis_logging(conf=None, **kwargs):
    logger = logging.getLogger('celery.task')
    redis_client = get_redis_client()
    logger.addHandler(RedisPubHandler(
        channel=LOG_KEY_PREFIX,
        redis_client=redis_client,
        level=logging.DEBUG))
    store_handler = RedisStoreHandler(
        prefix=LOG_KEY_PREFIX,
        redis_client=redis_client,
        level=logging.DEBUG)
    store_handler.setFormatter(logging.Formatter(
        fmt="%(levelname)s %(asctime)s %(module)s %(process)d %(message)s\n"))
    logger.addHandler(store_handler)


class RedisFormatter(logging.Formatter):
    def format(self, record):
        """
        JSON-encode a record for serializing through redis.

        Convert date to iso format, and stringify any exceptions.
        """
        task = get_current_task()
        if task and task.request:
            record.__dict__.update(task_id=task.request.id,
                                   task_name=task.name)
        else:
            raise UnknownTaskId()

        message = super(RedisFormatter, self).format(record)

        data = {
            'name': record.name,
            'level': record.levelno,
            'levelname': record.levelname,
            'filename': record.filename,
            'line_no': record.lineno,
            'message': message,
            'time': datetime.datetime.utcnow().isoformat(),
            'funcname': record.funcName,
            'traceback': record.exc_info,
            'task_id': record.task_id,
            'task_name': record.task_name,
        }

        # stringify exception data
        if record.exc_text:
            data['traceback'] = record.exc_text  # XXX should we format it?

        return json.dumps(data)


class RedisPubHandler(logging.Handler):
    """
    Publish messages to redis channel.
    """

    def __init__(self, channel, redis_client, *args, **kwargs):
        """
        Create a new logger for the given channel and redis_client.
        """
        super(RedisPubHandler, self).__init__(*args, **kwargs)
        self.channel = channel
        self.redis_client = redis_client
        self.formatter = RedisFormatter()

    def emit(self, record):
        """
        Publish record to redis logging channel
        """
        try:
            self.redis_client.publish(self.channel, self.format(record))
        except UnknownTaskId:
            pass


class RedisStoreHandler(logging.Handler):
    """
    Send logging messages to a redis store.
    """

    def __init__(self, prefix, redis_client, *args, **kwargs):
        """
        Create a new logger for the given channel and redis_client.
        """
        super(RedisStoreHandler, self).__init__(*args, **kwargs)
        self.prefix = prefix
        self.redis_client = redis_client

    def emit(self, record):
        """
        Publish record to redis logging channel
        """
        # See celery.app.log.TaskFormatter
        if record.task_id not in ('???', None):
            key = get_log_key(record.task_id)
            self.redis_client.append(key, self.format(record))



"""Helpers for managing logging to file in cubicweb-celerytask workers

Add this module 'cw_celerytask_helpers.filelogger' to CELERY_IMPORTS

You can control where logs are stored with the CUBICWEB_CELERYTASK_LOGDIR
config, the directory must exist and be writable.
"""
from __future__ import absolute_import

import logging
import gzip
import sys

import six
from celery import signals

from .fileutils import get_log_filename
from .fileutils import get_task_logs, flush_task_logs  # noqa

PY2 = sys.version_info[0] == 2
LOG_KEY_PREFIX = "cw:celerytask:log"


class UnknownTaskId(Exception):
    pass


def get_log_key(task_id):
    return "{0}:{1}".format(LOG_KEY_PREFIX, task_id)


@signals.celeryd_after_setup.connect
def setup_file_logging(conf=None, **kwargs):
    logger = logging.getLogger('celery.task')
    store_handler = FileStoreHandler(level=logging.DEBUG)
    store_handler.setFormatter(logging.Formatter(
        fmt="%(levelname)s %(asctime)s %(module)s %(process)d %(message)s\n"))
    logger.addHandler(store_handler)


class FileStoreHandler(logging.Handler):
    """
    Send logging messages to a file in
    """

    def emit(self, record):
        """
        Append logs to gzip log file in `self.logdir`
        """
        # See celery.app.log.TaskFormatter
        if record.task_id not in ('???', None):
            fname = get_log_filename(record.task_id)
            with gzip.open(fname, 'a') as f:
                formatted = self.format(record)
                if isinstance(formatted, six.text_type):
                    f.write(formatted.encode('utf-8'))
                else:
                    f.write(formatted)



import errno
import gzip
import os

import celery


def get_log_filename(task_id):
    logdir = celery.current_app.conf.get('CUBICWEB_CELERYTASK_LOGDIR')
    if not logdir:
        raise RuntimeError(
            "You asked for file-based log storage of the task logs "
            "but CUBICWEB_CELERYTASK_LOGDIR is not configured. "
            "Please set CUBICWEB_CELERYTASK_LOGDIR in your "
            "celery configuration.")
    return os.path.join(logdir, 'celerytask-{}.log.gz'.format(task_id))


def get_task_logs(task_id):
    """
    Get task logs by id
    """
    try:
        with gzip.open(get_log_filename(task_id), 'rb') as f:
            return f.read()
    except IOError as exc:
        if exc.errno != errno.ENOENT:
            raise
        return None


def flush_task_logs(task_id):
    """Delete task logs"""
    try:
        os.unlink(get_log_filename(task_id))
    except OSError as exc:
        if exc.errno != errno.ENOENT:
            raise



"""cw-celerytask-helpers packaging information"""

modname = 'cw_celerytask_helpers'
distname = 'cw-celerytask-helpers'

numversion = (0, 11, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'Worker side helpers for cubicweb-celerytask'
web = 'https://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'celery': '>=4,<5',
    'redis': None,
    'six': None,
}

classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
]



import os

import boto3
import celery


def get_bucket_name():
    bucket = celery.current_app.conf.get('CUBICWEB_CELERYTASK_S3_BUCKET')
    if not bucket:
        bucket = os.getenv('AWS_S3_BUCKET_NAME')
    return bucket


def get_key_name(task_id):
    key_pattern = celery.current_app.conf.get(
        'CUBICWEB_CELERYTASK_S3_KEY_PATTERN', 'celerytask-%s')
    return key_pattern % task_id


def get_s3_client():
    s3_endpoint_url = os.getenv(
        'AWS_S3_ENDPOINT_URL', 'https://s3.amazonaws.com')
    return boto3.client('s3', endpoint_url=s3_endpoint_url)


def get_task_logs(task_id):
    bucket = get_bucket_name()
    if not bucket:
        return None
    key = get_key_name(task_id)
    client = get_s3_client()
    try:
        result = client.get_object(Bucket=bucket, Key=key)
        return result['Body'].read()
    except client.exceptions.NoSuchKey:
        return None


def flush_task_logs(task_id):
    bucket = get_bucket_name()
    if not bucket:
        return
    key = get_key_name(task_id)
    client = get_s3_client()
    try:
        client.delete_object(Bucket=bucket, Key=key)
    except client.exceptions.NoSuchKey:
        pass



"""Helpers for cubicweb-celerytask workers

Add this module 'cw_celerytask_helpers.helpers' to CELERY_IMPORTS
"""

from __future__ import absolute_import

from . import redislogger  # noqa
from . import monitor  # noqa



"""Helpers for managing logging to CloudWatch in cubicweb-celerytask workers

Add this module 'cw_celerytask_helpers.cloudwatchlogger' to CELERY_IMPORTS
"""
from __future__ import absolute_import

import logging
from os import getenv
from signal import Signals

import celery
from celery import signals
from watchtower import CloudWatchLogHandler
import boto3

from .cloudwatchutils import get_stream_name
from .cloudwatchutils import get_task_logs, flush_task_logs  # noqa


@signals.task_prerun.connect
def setup_logging(conf=None, **kwargs):
    log_group = celery.current_app.conf.get('CUBICWEB_CELERYTASK_LOG_GROUP')
    if not log_group:
        raise RuntimeError(
            "You asked for CloudWatch-based log storage of the task logs "
            "but CUBICWEB_CELERYTASK_LOG_GROUP is not configured. "
            "Please set CUBICWEB_CELERYTASK_LOG_GROUP in your "
            "celery configuration.")
    task_id = kwargs.get('task_id')
    if task_id in ('???', None):
        return
    stream_name = get_stream_name(task_id)
    cloudwatch_endpoint_url = getenv('AWS_CLOUDWATCH_ENDPOINT_URL')
    handler = CloudWatchLogHandler(
        level=logging.DEBUG, log_group=log_group, stream_name=stream_name,
        endpoint_url=cloudwatch_endpoint_url, use_queues=False)
    handler.setFormatter(logging.Formatter(
        fmt="%(levelname)s %(asctime)s %(module)s %(process)d %(message)s\n"))
    logger = logging.getLogger('celery.task')
    logger.addHandler(handler)


@signals.task_postrun.connect
def uninstall_logging(conf=None, **kwargs):
    task_id = kwargs.get('task_id')
    if task_id in ('???', None):
        return
    logger = logging.getLogger('celery.task')
    log_group = celery.current_app.conf.get('CUBICWEB_CELERYTASK_LOG_GROUP')
    delete_stream = bool(celery.current_app.conf.get(
        'CUBICWEB_CELERYTASK_DELETE_LOG_STREAM'))
    stream_name = get_stream_name(task_id)
    for handler in logger.handlers:
        if isinstance(handler, CloudWatchLogHandler):
            logger.removeHandler(handler)
            if delete_stream:
                handler.cwl_client.delete_log_stream(
                    logGroupName=log_group,
                    logStreamName=stream_name)


@signals.task_revoked.connect
def delete_revoked_log_stream(conf=None, **kwargs):
    """executed on main celery process, hence logging handler is destroyed"""
    if 'signum' in kwargs and not isinstance(kwargs['signum'], Signals):
        # workaround for revoked task signal executed twice
        return
    request = kwargs.get('request')
    if not request:
        return
    task_id = request.get('id')
    if task_id in ('???', None):
        return
    delete_stream = bool(celery.current_app.conf.get(
        'CUBICWEB_CELERYTASK_DELETE_LOG_STREAM'))
    if not delete_stream:
        return
    log_group = celery.current_app.conf.get('CUBICWEB_CELERYTASK_LOG_GROUP')
    stream_name = get_stream_name(task_id)
    cloudwatch_endpoint_url = getenv('AWS_CLOUDWATCH_ENDPOINT_URL')
    client = boto3.client('logs', endpoint_url=cloudwatch_endpoint_url)
    client.delete_log_stream(logGroupName=log_group, logStreamName=stream_name)



import os.path
from setuptools import setup, find_packages

here = os.path.dirname(__file__)
# load metadata from the __pkginfo__.py file so there is no risk of conflict
# see https://packaging.python.org/en/latest/single_source_version.html
pkginfo = os.path.join(here, 'cw_celerytask_helpers', '__pkginfo__.py')
__pkginfo__ = {'__file__': pkginfo}
with open(pkginfo) as f:
    exec(f.read(), __pkginfo__)

setup(
    name=__pkginfo__['distname'],
    version=__pkginfo__['version'],
    description=__pkginfo__['description'],
    author=__pkginfo__['author'],
    author_email=__pkginfo__['author_email'],
    license=__pkginfo__['license'],
    classifiers=__pkginfo__['classifiers'],
    packages=find_packages('.'),
    install_requires=["{0} {1}".format(d, v and v or "").strip()
                      for d, v in __pkginfo__['__depends__'].items()],
    extras_require={
        'CloudWatch logging support': 'watchtower',
        'S3 logging support': 'boto3',
    },
)



"""Helpers for monitoring tasks in cubicweb-celerytask workers

Add this module 'cw_celerytask_helpers.monitor' to CELERY_IMPORTS
"""
from __future__ import absolute_import

import json

import celery
import celery.signals

from .redisutils import get_redis_client

MONITOR_KEY = "cw:celerytask:monitor"


def monitor_task(task_id, task_name):
    if celery.current_app.conf.get('CUBICWEB_CELERYTASK_MONITOR', True):
        client = get_redis_client()
        if client is not None:
            client.lpush(MONITOR_KEY, json.dumps({'task_id': task_id,
                                                  'task_name': task_name}))


@celery.signals.task_prerun.connect
def prerun(task_id, task, *args, **kwargs):
    monitor_task(task_id, task.name)


@celery.signals.task_postrun.connect
def postrun(task_id, task, *args, **kwargs):
    monitor_task(task_id, task.name)


@celery.signals.task_revoked.connect
def task_revoked(request, *args, **kwargs):
    monitor_task(request.task_id, request.task.name)



"""Helpers for managing logging to S3 in cubicweb-celerytask workers

Add this module 'cw_celerytask_helpers.s3logger' to CELERY_IMPORTS
"""
from __future__ import absolute_import

import os
import logging
from signal import Signals

import celery
import six
from celery import signals

from .s3utils import get_bucket_name, get_key_name, get_s3_client
from .s3utils import get_task_logs, flush_task_logs  # noqa


@signals.task_prerun.connect
def setup_logging(conf=None, **kwargs):
    bucket = get_bucket_name()
    if not bucket:
        raise RuntimeError(
            "You asked for S3-based log storage of the task logs "
            "but CUBICWEB_CELERYTASK_S3_BUCKET is not configured. "
            "Please either set CUBICWEB_CELERYTASK_S3_BUCKET in your "
            "celery configuration, or set the AWS_S3_BUCKET_NAME "
            "environment variable.")
    task_id = kwargs.get('task_id')
    if task_id in ('???', None):
        return
    key = get_key_name(task_id)
    handler = S3Handler(level=logging.DEBUG, bucket=bucket, key=key,
                        task_id=task_id)
    handler.setFormatter(logging.Formatter(
        fmt="%(levelname)s %(asctime)s %(module)s %(process)d %(message)s\n"))
    logger = logging.getLogger('celery.task')
    logger.addHandler(handler)


@signals.task_postrun.connect
def uninstall_logging(conf=None, **kwargs):
    task_id = kwargs.get('task_id')
    if task_id in ('???', None):
        return
    logger = logging.getLogger('celery.task')
    for handler in logger.handlers:
        if isinstance(handler, S3Handler):
            logger.removeHandler(handler)
            handler.send()


@signals.task_revoked.connect
def send_revoked_log(conf=None, **kwargs):
    """executed on main celery process, hence logging handler is destroyed"""
    if 'signum' in kwargs and not isinstance(kwargs['signum'], Signals):
        # workaround for revoked task signal executed twice
        return
    request = kwargs.get('request')
    if not request:
        return
    task_id = request.get('id')
    if task_id in ('???', None):
        return
    bucket = get_bucket_name()
    if not bucket:
        return
    key = get_key_name(task_id)
    fname = get_log_file_name(task_id)
    if not os.path.exists(fname):
        return
    try:
        client = get_s3_client()
        with open(fname, 'rb') as fobj:
            client.upload_fileobj(fobj, bucket, key)
    finally:
        os.unlink(fname)


class S3Handler(logging.Handler):
    def __init__(self, *args, **kwargs):
        self.key = kwargs.pop('key')
        self.bucket = kwargs.pop('bucket')
        self.task_id = kwargs.pop('task_id')
        self.closed = False
        self.fobj = open(get_log_file_name(self.task_id), 'w+b')
        self.s3_cnx = get_s3_client()
        super(S3Handler, self).__init__(*args, **kwargs)

    def emit(self, record):
        if self.closed or record.task_id in ('???', None):
            return
        formatted = self.format(record)
        if isinstance(formatted, six.text_type):
            self.fobj.write(formatted.encode('utf-8'))
        else:
            self.fobj.write(formatted)

    def send(self):
        self.closed = True
        try:
            self.fobj.seek(0)
            self.s3_cnx.upload_fileobj(self.fobj, self.bucket, self.key)
            self.fobj.close()
        finally:
            os.unlink(self.fobj.name)


def get_log_file_name(task_id):
    tmpdir = celery.current_app.conf.get(
        'CUBICWEB_CELERYTASK_S3_TMPDIR', '/tmp')
    return '%s/%s' % (tmpdir, task_id)



from time import time
from os import getenv

import boto3
import celery


def get_stream_name(task_id):
    stream_pattern = celery.current_app.conf.get(
        'CUBICWEB_CELERYTASK_STREAM_PATTERN', 'celerytask-%s')
    return stream_pattern % task_id


def get_logs_client():
    endpoint_url = getenv('AWS_CLOUDWATCH_ENDPOINT_URL')
    return boto3.client('logs', endpoint_url=endpoint_url)


def get_task_logs(task_id):
    log_group = celery.current_app.conf.get('CUBICWEB_CELERYTASK_LOG_GROUP')
    if not log_group:
        return None
    stream_name = get_stream_name(task_id)
    client = get_logs_client()
    try:
        logs = b''
        kwargs = {
            'logGroupName': log_group,
            'logStreamName': stream_name,
            'startFromHead': True,
            'endTime': int(time() * 1000),  # avoid getting newer logs
        }
        while True:
            result = client.get_log_events(**kwargs)
            for event in result['events']:
                message = event['message']
                if isinstance(message, str):
                    message = message.encode('utf-8')
                logs += message + b'\n'
            if 'nextForwardToken' not in result or (
                    'nextToken' in kwargs
                    and result['nextForwardToken'] == kwargs['nextToken']):
                return logs
            kwargs['nextToken'] = result['nextForwardToken']
    except client.exceptions.ResourceNotFoundException:
        return None


def flush_task_logs(task_id):
    log_group = celery.current_app.conf.get('CUBICWEB_CELERYTASK_LOG_GROUP')
    if not log_group:
        return
    stream_name = get_stream_name(task_id)
    client = get_logs_client()
    try:
        client.delete_log_stream(
            logGroupName=log_group,
            logStreamName=stream_name)
    except client.exceptions.ResourceNotFoundException:
        pass



import celery
import redis
from redis.client import Redis

LOG_KEY_PREFIX = "cw:celerytask:log"


def get_log_key(task_id):
    return "{0}:{1}".format(LOG_KEY_PREFIX, task_id)


def get_redis_client():
    conf = celery.current_app.conf
    url = conf.get('CUBICWEB_CELERYTASK_REDIS_URL')
    BROKER_TRANSPORT_OPTIONS = conf.get('BROKER_TRANSPORT_OPTIONS')
    if (url and url.startswith('redis-sentinel://')
            and 'sentinels' in BROKER_TRANSPORT_OPTIONS):
        from redis.sentinel import Sentinel
        service_name = BROKER_TRANSPORT_OPTIONS.get('service_name', 'master')
        socket_timeout = BROKER_TRANSPORT_OPTIONS.get('socket_timeout', 3)
        return Sentinel(BROKER_TRANSPORT_OPTIONS['sentinels'],
                        socket_timeout=socket_timeout).master_for(
                            service_name, redis_class=Redis,
                            socket_timeout=socket_timeout)
    elif url:
        return redis.Redis.from_url(url)


def get_task_logs(task_id):
    """
    Get task logs by id
    """
    redis_client = get_redis_client()
    return redis_client.get(get_log_key(task_id))


def flush_task_logs(task_id):
    """Delete task logs"""
    redis_client = get_redis_client()
    return redis_client.delete(get_log_key(task_id))
